# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import exceptions

from senlinclient.common import exc
from senlinclient import plugin


class Client(object):
    def __init__(self, prof=None, user_agent=None, **kwargs):
        try:
            conn = plugin.create_connection(prof=prof,
                                            user_agent=user_agent,
                                            **kwargs)
        except exceptions.HttpException as ex:
            exc.parse_exception(ex.details)

        self.conn = conn
        self.service = self.conn.cluster

    ######################################################################
    # The following operations are interfaces exposed to other software
    # which invokes senlinclient today.
    # These methods form a temporary translation layer. This layer will be
    # useless when OpenStackSDK has been adopted all senlin resources.
    ######################################################################

    def profile_types(self, **query):
        """List profile types

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-profile-types
        """
        return self.service.profile_types(**query)

    def get_profile_type(self, profile_type):
        """Show profile type details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #show-profile-type-details
        """
        return self.service.get_profile_type(profile_type)

    def list_profile_type_operations(self, profile_type):
        """Show profile type operations

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #show-profile-type-details
        """
        return self.service.list_profile_type_operations(profile_type)

    def profiles(self, **query):
        """List profiles

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-profiles
        """
        return self.service.profiles(**query)

    def create_profile(self, **attrs):
        """Create a profile

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#create-profile
        """
        return self.service.create_profile(**attrs)

    def get_profile(self, profile):
        """Show profile details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#show-profile-details
        """
        return self.service.get_profile(profile)

    def update_profile(self, profile, **attrs):
        """Update a profile

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#update-profile
        """
        return self.service.update_profile(profile, **attrs)

    def delete_profile(self, profile, ignore_missing=True):
        """Delete a profile

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#delete-profile
        """
        return self.service.delete_profile(profile, ignore_missing)

    def validate_profile(self, **attrs):
        """Validate a profile spec

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#validate-profile
        """
        return self.service.validate_profile(**attrs)

    def policy_types(self, **query):
        """List policy types

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-policy-types
        """
        return self.service.policy_types(**query)

    def get_policy_type(self, policy_type):
        """Show policy type details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #show-policy-type-details
        """
        return self.service.get_policy_type(policy_type)

    def policies(self, **query):
        """List policies

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-policies
        """
        return self.service.policies(**query)

    def create_policy(self, **attrs):
        """Create a policy

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#create-policy
        """
        return self.service.create_policy(**attrs)

    def get_policy(self, policy):
        """Show policy details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#show-policy-details
        """
        return self.service.get_policy(policy)

    def update_policy(self, policy, **attrs):
        """Update policy

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#update-policy
        """
        return self.service.update_policy(policy, **attrs)

    def delete_policy(self, policy, ignore_missing=True):
        """Delete policy

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#delete-policy
        """
        return self.service.delete_policy(policy, ignore_missing)

    def validate_policy(self, **attrs):
        """validate a policy spec

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#validate-policy
        """
        return self.service.validate_policy(**attrs)

    def clusters(self, **queries):
        """List clusters

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-clusters
        """
        return self.service.clusters(**queries)

    def create_cluster(self, **attrs):
        """Create a cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#create-cluster
        """
        return self.service.create_cluster(**attrs)

    def get_cluster(self, cluster):
        """Show cluster details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#show-cluster-details
        """
        return self.service.get_cluster(cluster)

    def update_cluster(self, cluster, **attrs):
        """Update cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#update-cluster
        """
        return self.service.update_cluster(cluster, **attrs)

    def delete_cluster(self, cluster, ignore_missing=True, force_delete=False):
        """Delete cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#delete-cluster
        """
        return self.service.delete_cluster(cluster, ignore_missing,
                                           force_delete)

    def cluster_add_nodes(self, cluster, nodes):
        """Add a node to cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #add-nodes-to-a-cluster
        """
        return self.service.add_nodes_to_cluster(cluster, nodes)

    def cluster_del_nodes(self, cluster, nodes):
        """Delete a node belongs to cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #remove-nodes-from-a-cluster
        """
        return self.service.remove_nodes_from_cluster(cluster, nodes)

    def cluster_replace_nodes(self, cluster, nodes):
        """Replace the nodes in a cluster with specified nodes

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #replace-nodes-in-a-cluster
        """
        return self.service.replace_nodes_in_cluster(cluster, nodes)

    def cluster_resize(self, cluster, **params):
        """Resize cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#resize-a-cluster
        """
        return self.service.resize_cluster(cluster, **params)

    def cluster_scale_out(self, cluster, count):
        """Scale out cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#scale-out-a-cluster
        """
        return self.service.scale_out_cluster(cluster, count)

    def cluster_scale_in(self, cluster, count):
        """Scale in cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#scale-in-a-cluster
        """
        return self.service.scale_in_cluster(cluster, count)

    def cluster_policies(self, cluster, **queries):
        """List all policies attached to cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #list-all-cluster-policies
        """
        return self.service.cluster_policies(cluster, **queries)

    def get_cluster_policy(self, policy, cluster):
        """Show details of a policy attached to cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #show-cluster-policy-details
        """
        return self.service.get_cluster_policy(policy, cluster)

    def cluster_attach_policy(self, cluster, policy, **attrs):
        """Attach a policy to cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #attach-a-policy-to-a-cluster
        """
        return self.service.attach_policy_to_cluster(cluster, policy, **attrs)

    def cluster_detach_policy(self, cluster, policy):
        """Detach a policy from cluster

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #detach-a-policy-from-a-cluster
        """
        return self.service.detach_policy_from_cluster(cluster, policy)

    def cluster_update_policy(self, cluster, policy, **attrs):
        """Update the policy attachment

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #update-a-policy-on-a-cluster
        """
        return self.service.update_cluster_policy(cluster, policy, **attrs)

    def collect_cluster_attrs(self, cluster, path):
        """Collect cluster attributes

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #collect-attributes-across-a-cluster
        """
        return self.service.collect_cluster_attrs(cluster, path)

    def check_cluster(self, cluster, **params):
        """Check cluster's health status

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #check-a-cluster-s-health-status
        """
        return self.service.check_cluster(cluster, **params)

    def recover_cluster(self, cluster, **params):
        """Recover cluster from failure state

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #recover-a-cluster-to-a-healthy-status
        """
        return self.service.recover_cluster(cluster, **params)

    def perform_operation_on_cluster(self, cluster, operation, **params):
        """Perform an operation on a cluster.

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #perform-an-operation-on-a-cluster
        """
        return self.service.perform_operation_on_cluster(cluster, operation,
                                                         **params)

    def nodes(self, **queries):
        """List nodes

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-nodes
        """
        return self.service.nodes(**queries)

    def create_node(self, **attrs):
        """Create a node

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#create-node
        """
        return self.service.create_node(**attrs)

    def adopt_node(self, preview=False, **attrs):
        """Adopt a node

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#adopt-node
        https://docs.openstack.org/api-ref/clustering/#adopt-node-preview
        """
        return self.service.adopt_node(preview, **attrs)

    def get_node(self, node, details=False):
        """Show node details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#show-node-details
        """
        return self.service.get_node(node, details=details)

    def update_node(self, node, **attrs):
        """Update node

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#update-node
        """
        return self.service.update_node(node, **attrs)

    def delete_node(self, node, ignore_missing=True, force_delete=False):
        """Delete node

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#delete-node
        """
        return self.service.delete_node(node, ignore_missing, force_delete)

    def check_node(self, node, **params):
        """Check node's health status

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#check-a-node-s-health
        """
        return self.service.check_node(node, **params)

    def recover_node(self, node, **params):
        """Recover node from failure state

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #recover-a-node-to-healthy-status
        """
        return self.service.recover_node(node, **params)

    def perform_operation_on_node(self, node, operation, **params):
        """Perform an operation on a node.

        Doc link:
        https://docs.openstack.org/api-ref/clustering/
        #perform-an-operation-on-a-node
        """
        return self.service.perform_operation_on_node(node, operation,
                                                      **params)

    def receivers(self, **queries):
        """List receivers

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-receivers
        """
        return self.service.receivers(**queries)

    def create_receiver(self, **attrs):
        """Creare a receiver

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#create-receiver
        """
        return self.service.create_receiver(**attrs)

    def get_receiver(self, receiver):
        """Show receiver details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#show-receiver-details
        """
        return self.service.get_receiver(receiver)

    def update_receiver(self, receiver, **attrs):
        """Update receiver

        Doc link:
        https://docs.openstack.org/api-ref-clustering-v1.html#updateReceiver
        """
        return self.service.update_receiver(receiver, **attrs)

    def delete_receiver(self, receiver, ignore_missing=True):
        """Delete receiver

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#delete-receiver
        """
        return self.service.delete_receiver(receiver, ignore_missing)

    def events(self, **queries):
        """List events

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-events
        """
        return self.service.events(**queries)

    def get_event(self, event):
        """Show event details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#shows-event-details
        """
        return self.service.get_event(event)

    def actions(self, **queries):
        """List actions

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-actions
        """
        return self.service.actions(**queries)

    def get_action(self, action):
        """Show action details

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#show-action-details
        """
        return self.service.get_action(action)

    def update_action(self, action, **attrs):
        """Update an action

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#update-action
        """
        return self.service.update_action(action, **attrs)

    def services(self, **queries):
        """List services

        Doc link:
        https://docs.openstack.org/api-ref/clustering/#list-services
        """
        return self.service.services(**queries)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 cluster policy action implementations"""

import logging

from osc_lib.command import command
from osc_lib import utils
from oslo_utils import strutils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ClusterPolicyList(command.Lister):
    """List policies from cluster."""

    log = logging.getLogger(__name__ + ".ClusterPolicyList")

    def get_parser(self, prog_name):
        parser = super(ClusterPolicyList, self).get_parser(prog_name)
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned results. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter "
                   "keys are: ['is_enabled', 'policy_type', 'policy_name']"),
            action='append'
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc).  The valid "
                   "sort keys are: ['enabled']")
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to query on')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        columns = ['policy_id', 'policy_name', 'policy_type', 'is_enabled']
        cluster = senlin_client.get_cluster(parsed_args.cluster)
        queries = {
            'sort': parsed_args.sort,
        }

        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        policies = senlin_client.cluster_policies(cluster.id, **queries)
        formatters = {}
        if not parsed_args.full_id:
            formatters = {
                'policy_id': lambda x: x[:8]
            }
        return (
            columns,
            (utils.get_item_properties(p, columns,
                                       formatters=formatters)
             for p in policies)
        )


class ClusterPolicyShow(command.ShowOne):
    """Show a specific policy that is bound to the specified cluster."""

    log = logging.getLogger(__name__ + ".ClusterPolicyShow")

    def get_parser(self, prog_name):
        parser = super(ClusterPolicyShow, self).get_parser(prog_name)
        parser.add_argument(
            '--policy',
            metavar='<policy>',
            required=True,
            help=_('ID or name of the policy to query on')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('ID or name of the cluster to query on')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        policy = senlin_client.get_cluster_policy(parsed_args.policy,
                                                  parsed_args.cluster)
        data = policy.to_dict()
        columns = sorted(data.keys())
        return columns, utils.get_dict_properties(data, columns)


class ClusterPolicyUpdate(command.Command):
    """Update a policy's properties on a cluster."""

    log = logging.getLogger(__name__ + ".ClusterPolicyUpdate")

    def get_parser(self, prog_name):
        parser = super(ClusterPolicyUpdate, self).get_parser(prog_name)
        parser.add_argument(
            '--policy',
            metavar='<policy>',
            required=True,
            help=_('ID or name of policy to be updated')
        )
        parser.add_argument(
            '--enabled',
            metavar='<boolean>',
            required=True,
            help=_('Whether the policy should be enabled')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        kwargs = {
            'enabled': strutils.bool_from_string(parsed_args.enabled,
                                                 strict=True),
        }

        resp = senlin_client.update_cluster_policy(parsed_args.cluster,
                                                   parsed_args.policy,
                                                   **kwargs)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
        else:
            print('Request error: %s' % resp)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import cluster_policy as osc_cluster_policy


class TestClusterPolicy(fakes.TestClusteringv1):
    def setUp(self):
        super(TestClusterPolicy, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestClusterPolicyList(TestClusterPolicy):

    def setUp(self):
        super(TestClusterPolicyList, self).setUp()
        self.cmd = osc_cluster_policy.ClusterPolicyList(self.app, None)
        fake_cluster = mock.Mock(id='C1')
        self.mock_client.get_cluster = mock.Mock(return_value=fake_cluster)
        fake_binding = mock.Mock(
            cluster_id="7d85f602-a948-4a30-afd4-e84f47471c15",
            cluster_name="my_cluster",
            is_enabled=True,
            id="06be3a1f-b238-4a96-a737-ceec5714087e",
            policy_id="714fe676-a08f-4196-b7af-61d52eeded15",
            policy_name="my_policy",
            policy_type="senlin.policy.deletion-1.0"
        )
        fake_binding.to_dict = mock.Mock(return_value={})
        self.mock_client.cluster_policies = mock.Mock(
            return_value=[fake_binding])

    def test_cluster_policy_list(self):
        arglist = ['--sort', 'name:asc', '--filter', 'name=my_policy',
                   'my_cluster', '--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        expected_columns = ['policy_id', 'policy_name', 'policy_type',
                            'is_enabled']

        columns, data = self.cmd.take_action(parsed_args)

        self.mock_client.get_cluster.assert_called_with('my_cluster')
        self.mock_client.cluster_policies.assert_called_with(
            'C1', name="my_policy", sort="name:asc")
        self.assertEqual(expected_columns, columns)


class TestClusterPolicyShow(TestClusterPolicy):

    def setUp(self):
        super(TestClusterPolicyShow, self).setUp()
        self.cmd = osc_cluster_policy.ClusterPolicyShow(self.app, None)
        fake_binding = mock.Mock(
            cluster_id="7d85f602-a948-4a30-afd4-e84f47471c15",
            cluster_name="my_cluster",
            is_enabled=True,
            id="06be3a1f-b238-4a96-a737-ceec5714087e",
            policy_id="714fe676-a08f-4196-b7af-61d52eeded15",
            policy_name="my_policy",
            policy_type="senlin.policy.deletion-1.0"
        )
        fake_binding.to_dict = mock.Mock(return_value={})
        self.mock_client.get_cluster_policy = mock.Mock(
            return_value=fake_binding)

    def test_cluster_policy_show(self):
        arglist = ['--policy', 'my_policy', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_cluster_policy.assert_called_with('my_policy',
                                                               'my_cluster')


class TestClusterPolicyUpdate(TestClusterPolicy):

    def setUp(self):
        super(TestClusterPolicyUpdate, self).setUp()
        self.cmd = osc_cluster_policy.ClusterPolicyUpdate(self.app, None)
        fake_resp = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}
        self.mock_client.update_cluster_policy = mock.Mock(
            return_value=fake_resp)

    def test_cluster_policy_update(self):
        arglist = ['--policy', 'my_policy', '--enabled', 'true', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.update_cluster_policy.assert_called_with(
            'my_cluster', 'my_policy', enabled=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.exceptions import base as kae_base
from keystoneauth1.exceptions import http as kae_http
from openstack import exceptions as sdkexc
from oslo_serialization import jsonutils
from requests import exceptions as reqexc

from senlinclient.common.i18n import _

verbose = False


class BaseException(Exception):
    """An error occurred."""
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return self.message or self.__class__.__doc__


class CommandError(BaseException):
    """Invalid usage of CLI."""


class FileFormatError(BaseException):
    """Illegal file format detected."""


class HTTPException(BaseException):
    """Base exception for all HTTP-derived exceptions."""
    code = 'N/A'

    def __init__(self, error=None):
        super(HTTPException, self).__init__(error)
        try:
            self.error = error
            if 'error' not in self.error:
                raise KeyError(_('Key "error" not exists'))
        except KeyError:
            # If key 'error' does not exist, self.message becomes
            # no sense. In this case, we return doc of current
            # exception class instead.
            self.error = {'error': {'message': self.__class__.__doc__}}
        except Exception:
            self.error = {'error':
                          {'message': self.message or self.__class__.__doc__}}

    def __str__(self):
        message = self.error['error'].get('message', 'Internal Error')
        if verbose:
            traceback = self.error['error'].get('traceback', '')
            return (_('ERROR: %(message)s\n%(traceback)s') %
                    {'message': message, 'traceback': traceback})
        else:
            code = self.error['error'].get('code', 'Unknown')
            return _('ERROR(%(code)s): %(message)s') % {'code': code,
                                                        'message': message}


class ClientError(HTTPException):
    pass


class ServerError(HTTPException):
    pass


class HTTPBadRequest(ClientError):
    # 400
    pass


class HTTPUnauthorized(ClientError):
    # 401
    pass


class HTTPForbidden(ClientError):
    # 403
    pass


class HTTPNotFound(ClientError):
    # 404
    pass


class HTTPMethodNotAllowed(ClientError):
    # 405
    pass


class HTTPNotAcceptable(ClientError):
    # 406
    pass


class HTTPProxyAuthenticationRequired(ClientError):
    # 407
    pass


class HTTPRequestTimeout(ClientError):
    # 408
    pass


class HTTPConflict(ClientError):
    # 409
    pass


class HTTPGone(ClientError):
    # 410
    pass


class HTTPLengthRequired(ClientError):
    # 411
    pass


class HTTPPreconditionFailed(ClientError):
    # 412
    pass


class HTTPRequestEntityTooLarge(ClientError):
    # 413
    pass


class HTTPRequestURITooLong(ClientError):
    # 414
    pass


class HTTPUnsupportedMediaType(ClientError):
    # 415
    pass


class HTTPRequestRangeNotSatisfiable(ClientError):
    # 416
    pass


class HTTPExpectationFailed(ClientError):
    # 417
    pass


class HTTPInternalServerError(ServerError):
    # 500
    pass


class HTTPNotImplemented(ServerError):
    # 501
    pass


class HTTPBadGateway(ServerError):
    # 502
    pass


class HTTPServiceUnavailable(ServerError):
    # 503
    pass


class HTTPGatewayTimeout(ServerError):
    # 504
    pass


class HTTPVersionNotSupported(ServerError):
    # 505
    pass


class ConnectionRefused(HTTPException):
    # 111
    pass


_EXCEPTION_MAP = {
    111: ConnectionRefused,
    400: HTTPBadRequest,
    401: HTTPUnauthorized,
    403: HTTPForbidden,
    404: HTTPNotFound,
    405: HTTPMethodNotAllowed,
    406: HTTPNotAcceptable,
    407: HTTPProxyAuthenticationRequired,
    408: HTTPRequestTimeout,
    409: HTTPConflict,
    410: HTTPGone,
    411: HTTPLengthRequired,
    412: HTTPPreconditionFailed,
    413: HTTPRequestEntityTooLarge,
    414: HTTPRequestURITooLong,
    415: HTTPUnsupportedMediaType,
    416: HTTPRequestRangeNotSatisfiable,
    417: HTTPExpectationFailed,
    500: HTTPInternalServerError,
    501: HTTPNotImplemented,
    502: HTTPBadGateway,
    503: HTTPServiceUnavailable,
    504: HTTPGatewayTimeout,
    505: HTTPVersionNotSupported,
}


def parse_exception(exc):
    """Parse exception code and yield useful information.

    :param exc: details of the exception.
    """
    if isinstance(exc, sdkexc.HttpException):
        if exc.details is None:
            data = exc.response.json()
            code = data.get('code', None)
            message = data.get('message', None)
            error = data.get('error', None)
            if error:
                record = {
                    'error': {
                        'code': exc.http_status,
                        'message': message or exc.message
                    }
                }
            else:
                info = data.values()[0]
                record = {
                    'error': {
                        'code': info.get('code', code),
                        'message': info.get('message', message)
                    }
                }
        else:
            try:
                record = jsonutils.loads(exc.details)
            except Exception:
                # If the exc.details is not in JSON format
                record = {
                    'error': {
                        'code': exc.http_status,
                        'message': exc,
                    }
                }
    elif isinstance(exc, reqexc.RequestException):
        # Exceptions that are not captured by SDK
        record = {
            'error': {
                'code': exc.message[1].errno,
                'message': exc.message[0],
            }
        }

    elif isinstance(exc, str):
        record = jsonutils.loads(exc)
    # some exception from keystoneauth1 is not shaped by SDK
    elif isinstance(exc, kae_http.HttpError):
        record = {
            'error': {
                'code': exc.http_status,
                'message': exc.message
            }
        }
    elif isinstance(exc, kae_base.ClientException):
        record = {
            'error': {
                # other exceptions from keystoneauth1 is an internal
                # error to senlin, so set status code to 500
                'code': 500,
                'message': exc.message
            }
        }
    else:
        print(_('Unknown exception: %s') % exc)
        return

    try:
        code = record['error']['code']
    except KeyError as err:
        print(_('Malformed exception record, missing field "%s"') % err)
        print(_('Original error record: %s') % record)
        return

    if code in _EXCEPTION_MAP:
        inst = _EXCEPTION_MAP.get(code)
        raise inst(record)
    else:
        raise HTTPException(record)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 policy type action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc

from senlinclient.common import format_utils
from senlinclient.common.i18n import _


class PolicyTypeList(command.Lister):
    """List the available policy types."""

    log = logging.getLogger(__name__ + ".PolicyTypeList")

    def get_parser(self, prog_name):
        parser = super(PolicyTypeList, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        types = senlin_client.policy_types()
        columns = ['name', 'version', 'support_status']
        results = []
        for t in types:
            if t.support_status:
                for v in t.support_status.keys():
                    st_list = '\n'.join([
                        ' since '.join((item['status'], item['since']))
                        for item in t.support_status[v]
                    ])

                    results.append((t.name, v, st_list))
            else:
                columns = ['name', 'version']
                results.append((t.name.split('-')[0], t.name.split('-')[1]))

        return columns, sorted(results)


class PolicyTypeShow(format_utils.YamlFormat):
    """Get the details about a policy type."""

    log = logging.getLogger(__name__ + ".PolicyTypeShow")

    def get_parser(self, prog_name):
        parser = super(PolicyTypeShow, self).get_parser(prog_name)
        parser.add_argument(
            'type_name',
            metavar='<type-name>',
            help=_('Policy type to retrieve')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        try:
            res = senlin_client.get_policy_type(parsed_args.type_name)
        except sdk_exc.ResourceNotFound:
            raise exc.CommandError(_('Policy Type not found: %s')
                                   % parsed_args.type_name)
        data = res.to_dict()
        rows = data.values()
        columns = data.keys()
        return columns, rows



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class ProfileTypeTest(base.OpenStackClientTestBase):
    """Test for profile types

    Basic smoke test for the Openstack CLI commands which do not require
    creating or modifying.
    """
    def test_profile_type_list(self):
        result = self.openstack('cluster profile type list')
        profile_type = self.parser.listing(result)
        columns = ['name', 'version']
        if any('support_status' in i for i in profile_type):
            columns.append('support_status')
        self.assertTableStruct(profile_type, columns)

    def test_profile_list_debug(self):
        self.openstack('cluster profile type list', flags='--debug')

    def test_profile_type_show(self):
        params = ['container.dockerinc.docker-1.0',
                  'os.heat.stack-1.0',
                  'os.nova.server-1.0']
        for param in params:
            cmd = 'cluster profile type show %s' % param
            self.openstack(cmd)

    def test_profile_type_show_json(self):
        self.openstack('cluster profile type show os.nova.server-1.0 -f json')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from heatclient.common import template_utils
from unittest import mock

import testtools

from senlinclient.common import exc
from senlinclient.common.i18n import _
from senlinclient.common import utils


class UtilTest(testtools.TestCase):

    def test_format_parameter(self):
        params = ['status=ACTIVE;name=cluster1']
        format_params = {'status': 'ACTIVE', 'name': 'cluster1'}
        self.assertEqual(format_params,
                         utils.format_parameters(params))

    def test_format_parameter_split(self):
        params = ['status=ACTIVE', 'name=cluster1']
        format_params = {'status': 'ACTIVE', 'name': 'cluster1'}
        self.assertEqual(format_params,
                         utils.format_parameters(params))

    def test_format_parameter_none_dict(self):
        params = ['{}']
        self.assertEqual({}, utils.format_parameters(params))

    def test_format_parameter_none(self):
        self.assertEqual({}, utils.format_parameters(None))

    def test_format_parameter_bad_format(self):
        params = ['status:ACTIVE;name:cluster1']
        ex = self.assertRaises(exc.CommandError,
                               utils.format_parameters,
                               params)
        msg = _('Malformed parameter(status:ACTIVE). '
                'Use the key=value format.')
        self.assertEqual(msg, str(ex))

    @mock.patch.object(template_utils,
                       'process_multiple_environments_and_files')
    @mock.patch.object(template_utils, 'get_template_contents')
    def test_process_stack_spec(self, mock_get_temp, mock_process):
        spec = {
            'template': 'temp.yaml',
            'disable_rollback': True,
            'context': {
                'region_name': 'RegionOne'
            },
        }
        tpl_files = {'fake_key1': 'fake_value1'}
        template = mock.Mock()
        mock_get_temp.return_value = tpl_files, template
        env_files = {'fake_key2': 'fake_value2'}
        env = mock.Mock()
        mock_process.return_value = env_files, env
        new_spec = utils.process_stack_spec(spec)
        stack_spec = {
            'disable_rollback': True,
            'context': {
                'region_name': 'RegionOne',
            },
            'parameters': {},
            'timeout': 60,
            'template': template,
            'files': {
                'fake_key1': 'fake_value1',
                'fake_key2': 'fake_value2',
            },
            'environment': env
        }
        self.assertEqual(stack_spec, new_spec)
        mock_get_temp.assert_called_once_with(template_file='temp.yaml')
        mock_process.assert_called_once_with(env_paths=None)

    def test_json_formatter_with_empty_json(self):
        params = {}
        self.assertEqual('{}', utils.json_formatter(params))

    def test_list_formatter_with_list(self):
        params = ['foo', 'bar']
        self.assertEqual('foo\nbar', utils.list_formatter(params))

    def test_list_formatter_with_empty_list(self):
        params = []
        self.assertEqual('', utils.list_formatter(params))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 profile action implementations"""

import logging
import sys

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ShowProfile(command.ShowOne):
    """Show profile details."""

    log = logging.getLogger(__name__ + ".ShowProfile")

    def get_parser(self, prog_name):
        parser = super(ShowProfile, self).get_parser(prog_name)
        parser.add_argument(
            'profile',
            metavar='<profile>',
            help='Name or ID of profile to show',
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        return _show_profile(senlin_client, profile_id=parsed_args.profile)


def _show_profile(senlin_client, profile_id):
    try:
        data = senlin_client.get_profile(profile_id)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError('Profile not found: %s' % profile_id)
    else:
        formatters = {}
        formatters['metadata'] = senlin_utils.json_formatter
        formatters['spec'] = senlin_utils.nested_dict_formatter(
            ['type', 'version', 'properties'],
            ['property', 'value'])

        data = data.to_dict()
        columns = sorted(data.keys())
        return columns, utils.get_dict_properties(data, columns,
                                                  formatters=formatters)


class ListProfile(command.Lister):
    """List profiles that meet the criteria."""

    log = logging.getLogger(__name__ + ".ListProfile")

    def get_parser(self, prog_name):
        parser = super(ListProfile, self).get_parser(prog_name)
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of profiles returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return profiles that appear after the given profile '
                   'ID')
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of keys"
                   " separated by commas. Each key can be optionally appended "
                   "by a sort direction (:asc or :desc). The valid sort_keys "
                   "are:['type', 'name', 'created_at', 'updated_at']")
        )
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned profiles. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter "
                   "keys are: ['type', 'name']"),
            action='append'
        )
        parser.add_argument(
            '--global-project',
            default=False,
            action="store_true",
            help=_('Indicate that the list should include profiles from'
                   ' all projects. This option is subject to access policy '
                   'checking. Default is False')
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering

        columns = ['id', 'name', 'type', 'created_at']
        queries = {
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
            'sort': parsed_args.sort,
            'global_project': parsed_args.global_project,
        }
        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))
        data = senlin_client.profiles(**queries)

        formatters = {}
        if parsed_args.global_project:
            columns.append('project_id')
        if not parsed_args.full_id:
            formatters = {
                'id': lambda x: x[:8],
            }
            if 'project_id' in columns:
                formatters['project_id'] = lambda x: x[:8]

        return (
            columns,
            (utils.get_item_properties(p, columns, formatters=formatters)
             for p in data)
        )


class DeleteProfile(command.Command):
    """Delete profile(s)."""

    log = logging.getLogger(__name__ + ".DeleteProfile")

    def get_parser(self, prog_name):
        parser = super(DeleteProfile, self).get_parser(prog_name)
        parser.add_argument(
            'profile',
            metavar='<profile>',
            nargs='+',
            help=_('Name or ID of profile(s) to delete')
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help=_('Skip yes/no prompt (assume yes)')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        try:
            if not parsed_args.force and sys.stdin.isatty():
                sys.stdout.write(
                    _("Are you sure you want to delete this profile(s)"
                      " [y/N]?"))
                sys.stdout.flush()
                prompt_response = sys.stdin.readline().lower()
                if not prompt_response.startswith('y'):
                    return
        except KeyboardInterrupt:  # Ctrl-c
            self.log.info('Ctrl-c detected.')
            return
        except EOFError:  # Ctrl-d
            self.log.info('Ctrl-d detected')
            return

        failure_count = 0
        for pid in parsed_args.profile:
            try:
                senlin_client.delete_profile(pid, False)
            except Exception as ex:
                failure_count += 1
                print(ex)
        if failure_count:
            raise exc.CommandError(_('Failed to delete %(count)s of the '
                                     '%(total)s specified profile(s).') %
                                   {'count': failure_count,
                                   'total': len(parsed_args.profile)})
        print('Profile deleted: %s' % parsed_args.profile)


class CreateProfile(command.ShowOne):
    """Create a profile."""

    log = logging.getLogger(__name__ + ".CreateProfile")

    def get_parser(self, prog_name):
        parser = super(CreateProfile, self).get_parser(prog_name)
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_('Metadata values to be attached to the profile. '
                   'This can be specified multiple times, or once with '
                   'key-value pairs separated by a semicolon'),
            action='append'
        )
        parser.add_argument(
            '--spec-file',
            metavar='<spec-file>',
            required=True,
            help=_('The spec file used to create the profile')
        )
        parser.add_argument(
            'name',
            metavar='<profile-name>',
            help=_('Name of the profile to create')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        spec = senlin_utils.get_spec_content(parsed_args.spec_file)
        type_name = spec.get('type', None)
        type_version = spec.get('version', None)
        properties = spec.get('properties', None)
        if type_name is None:
            raise exc.CommandError(_("Missing 'type' key in spec file."))
        if type_version is None:
            raise exc.CommandError(_("Missing 'version' key in spec file."))
        if properties is None:
            raise exc.CommandError(_("Missing 'properties' key in spec file."))

        if type_name == 'os.heat.stack':
            stack_properties = senlin_utils.process_stack_spec(properties)
            spec['properties'] = stack_properties

        params = {
            'name': parsed_args.name,
            'spec': spec,
            'metadata': senlin_utils.format_parameters(parsed_args.metadata),
        }

        profile = senlin_client.create_profile(**params)
        return _show_profile(senlin_client, profile_id=profile.id)


class UpdateProfile(command.ShowOne):
    """Update a profile."""

    log = logging.getLogger(__name__ + ".UpdateProfile")

    def get_parser(self, prog_name):
        parser = super(UpdateProfile, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('The new name for the profile')
        )
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Metadata values to be attached to the profile. "
                   "This can be specified multiple times, or once with "
                   "key-value pairs separated by a semicolon. Use '{}' "
                   "can clean metadata "),
            action='append'
        )
        parser.add_argument(
            'profile',
            metavar='<profile>',
            help=_('Name or ID of the profile to update')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        params = {
            'name': parsed_args.name,
        }
        if parsed_args.metadata:
            params['metadata'] = senlin_utils.format_parameters(
                parsed_args.metadata)

        # Find the profile first, we need its id
        profile = senlin_client.find_profile(parsed_args.profile)
        if profile is None:
            raise exc.CommandError(_('Profile not found: %s') %
                                   parsed_args.profile)
        senlin_client.update_profile(profile.id, **params)
        return _show_profile(senlin_client, profile_id=profile.id)


class ValidateProfile(command.ShowOne):
    """Validate a profile."""

    log = logging.getLogger(__name__ + ".ValidateProfile")

    def get_parser(self, prog_name):
        parser = super(ValidateProfile, self).get_parser(prog_name)
        parser.add_argument(
            '--spec-file',
            metavar='<spec-file>',
            required=True,
            help=_('The spec file of the profile to be validated')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        spec = senlin_utils.get_spec_content(parsed_args.spec_file)
        type_name = spec.get('type', None)
        type_version = spec.get('version', None)
        properties = spec.get('properties', None)
        if type_name is None:
            raise exc.CommandError(_("Missing 'type' key in spec file."))
        if type_version is None:
            raise exc.CommandError(_("Missing 'version' key in spec file."))
        if properties is None:
            raise exc.CommandError(_("Missing 'properties' key in spec file."))

        if type_name == 'os.heat.stack':
            stack_properties = senlin_utils.process_stack_spec(properties)
            spec['properties'] = stack_properties

        params = {
            'spec': spec,
        }

        profile = senlin_client.validate_profile(**params)

        formatters = {}
        formatters['metadata'] = senlin_utils.json_formatter
        formatters['spec'] = senlin_utils.nested_dict_formatter(
            ['type', 'version', 'properties'],
            ['property', 'value'])

        columns = [
            'created_at',
            'domain',
            'id',
            'metadata',
            'name',
            'project_id',
            'spec',
            'type',
            'updated_at',
            'user_id'
        ]
        return columns, utils.get_dict_properties(profile.to_dict(), columns,
                                                  formatters=formatters)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from heatclient.common import template_utils
from oslo_serialization import jsonutils
from oslo_utils import importutils
import prettytable
import yaml

from senlinclient.common import exc
from senlinclient.common.i18n import _


def import_versioned_module(version, submodule=None):
    module = 'senlinclient.v%s' % version
    if submodule:
        module = '.'.join((module, submodule))
        return importutils.import_module(module)


def format_nested_dict(d, fields, column_names):
    if d is None:
        return ''
    pt = prettytable.PrettyTable(caching=False, print_empty=False,
                                 header=True, field_names=column_names)
    for n in column_names:
        pt.align[n] = 'l'

    keys = sorted(d.keys())
    for field in keys:
        value = d[field]
        if not isinstance(value, str):
            value = jsonutils.dumps(value, indent=2, ensure_ascii=False)
        if value is None:
            value = '-'
        pt.add_row([field, value.strip('"')])

    return pt.get_string()


def nested_dict_formatter(d, column_names):
    return lambda o: format_nested_dict(o, d, column_names)


def json_formatter(js):
    return jsonutils.dumps(js, indent=2, ensure_ascii=False)


def list_formatter(record):
    return '\n'.join(record or [])


def print_action_result(rid, res):
    if res[0] == "OK":
        output = _("accepted by action %s") % res[1]
    else:
        output = _("failed due to '%s'") % res[1]
    print(_(" %(cid)s: %(output)s") % {"cid": rid, "output": output})


def format_parameters(params, parse_semicolon=True):
    """Reformat parameters into dict of format expected by the API."""
    if not params or params == ['{}']:
        return {}

    if parse_semicolon:
        # expect multiple invocations of --parameters but fall back to ';'
        # delimited if only one --parameters is specified
        if len(params) == 1:
            params = params[0].split(';')

    parameters = {}
    for p in params:
        try:
            (n, v) = p.split(('='), 1)
        except ValueError:
            msg = _('Malformed parameter(%s). Use the key=value format.') % p
            raise exc.CommandError(msg)

        if n not in parameters:
            parameters[n] = v
        else:
            if not isinstance(parameters[n], list):
                parameters[n] = [parameters[n]]
            parameters[n].append(v)

    return parameters


def format_json_parameter(param):
    '''Return JSON dict from JSON formatted param.

    :parameter param  JSON formatted string
    :return JSON dict
    '''
    if not param:
        return {}

    try:
        return jsonutils.loads(param)
    except ValueError:
        msg = _('Malformed parameter(%s). Use the JSON format.') % param
        raise exc.CommandError(msg)


def get_spec_content(filename):
    with open(filename, 'r') as f:
        try:
            data = yaml.safe_load(f)
        except Exception as ex:
            raise exc.CommandError(_('The specified file is not a valid '
                                     'YAML file: %s') % str(ex))
    return data


def process_stack_spec(spec):
    # Heat stack is a headache, because it demands for client side file
    # content processing
    try:
        tmplfile = spec.get('template', None)
    except AttributeError as ex:
        raise exc.FileFormatError(_('The specified file is not a valid '
                                    'YAML file: %s') % str(ex))
    if not tmplfile:
        raise exc.FileFormatError(_('No template found in the given '
                                    'spec file'))

    tpl_files, template = template_utils.get_template_contents(
        template_file=tmplfile)

    env_files, env = template_utils.process_multiple_environments_and_files(
        env_paths=spec.get('environment', None))

    new_spec = {
        # TODO(Qiming): add context support
        'disable_rollback': spec.get('disable_rollback', True),
        'context': spec.get('context', {}),
        'parameters': spec.get('parameters', {}),
        'timeout': spec.get('timeout', 60),
        'template': template,
        'files': dict(list(tpl_files.items()) + list(env_files.items())),
        'environment': env
    }

    return new_spec



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import exceptions

from senlinclient.tests.functional import base


class FakeTest(base.OpenStackClientTestBase):
    """Test if fake actions can be detected"""

    def test_fake_action(self):
        self.assertRaises(exceptions.CommandFailed,
                          self.openstack,
                          'this-does-not-exist')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
from unittest import mock

from osc_lib.tests import utils
import requests

from oslo_serialization import jsonutils

AUTH_TOKEN = "foobar"
AUTH_URL = "http://0.0.0.0"
USERNAME = "itchy"
PASSWORD = "scratchy"

TEST_RESPONSE_DICT_V3 = {
    "token": {
        "audit_ids": [
            "a"
        ],
        "catalog": [
        ],
        "expires_at": "2034-09-29T18:27:15.978064Z",
        "extras": {},
        "issued_at": "2014-09-29T17:27:15.978097Z",
        "methods": [
            "password"
        ],
        "project": {
            "domain": {
                "id": "default",
                "name": "Default"
            },
            "id": "bbb",
            "name": "project"
        },
        "roles": [
        ],
        "user": {
            "domain": {
                "id": "default",
                "name": "Default"
            },
            "id": "aaa",
            "name": USERNAME
        }
    }
}
TEST_VERSIONS = {
    "versions": {
        "values": [
            {
                "id": "v3.0",
                "links": [
                    {
                        "href": AUTH_URL,
                        "rel": "self"
                    }
                ],
                "media-types": [
                    {
                        "base": "application/json",
                        "type": "application/vnd.openstack.identity-v3+json"
                    },
                    {
                        "base": "application/xml",
                        "type": "application/vnd.openstack.identity-v3+xml"
                    }
                ],
                "status": "stable",
                "updated": "2013-03-06T00:00:00Z"
            }
        ]
    }
}


class FakeStdout(object):
    def __init__(self):
        self.content = []

    def write(self, text):
        self.content.append(text)

    def make_string(self):
        result = ''
        for line in self.content:
            result = result + line
        return result


class FakeApp(object):
    def __init__(self, _stdout):
        self.stdout = _stdout
        self.client_manager = None
        self.stdin = sys.stdin
        self.stdout = _stdout or sys.stdout
        self.stderr = sys.stderr


class FakeClient(object):
    def __init__(self, **kwargs):
        self.auth_url = kwargs['auth_url']
        self.token = kwargs['token']


class FakeClientManager(object):
    def __init__(self):
        self.compute = None
        self.identity = None
        self.image = None
        self.object_store = None
        self.volume = None
        self.network = None
        self.session = None
        self.auth_ref = None


class FakeModule(object):
    def __init__(self, name, version):
        self.name = name
        self.__version__ = version


class FakeResource(object):
    def __init__(self, manager, info, loaded=False):
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._loaded = loaded

    def _add_details(self, info):
        for (k, v) in info.items():
            setattr(self, k, v)

    def __repr__(self):
        reprkeys = sorted(k for k in self.__dict__.keys() if k[0] != '_' and
                          k != 'manager')
        info = ", ".join("%s=%s" % (k, getattr(self, k)) for k in reprkeys)
        return "<%s %s>" % (self.__class__.__name__, info)


class FakeResponse(requests.Response):
    def __init__(self, headers={}, status_code=200, data=None, encoding=None):
        super(FakeResponse, self).__init__()

        self.status_code = status_code

        self.headers.update(headers)
        self._content = jsonutils.dump_as_bytes(data)


class FakeClusteringv1Client(object):
    def __init__(self, **kwargs):
        self.http_client = mock.Mock()
        self.http_client.auth_token = kwargs['token']
        self.profiles = FakeResource(None, {})


class TestClusteringv1(utils.TestCommand):
    def setUp(self):
        super(TestClusteringv1, self).setUp()

        self.app.client_manager.clustering = FakeClusteringv1Client(
            token=AUTH_TOKEN, auth_url=AUTH_URL
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class NodeTest(base.OpenStackClientTestBase):
    """Test for nodes"""

    def test_node_list(self):
        result = self.openstack('cluster node list')
        node_list = self.parser.listing(result)
        self.assertTableStruct(node_list, ['id', 'name', 'index', 'status',
                                           'cluster_id', 'physical_id',
                                           'profile_name', 'created_at',
                                           'updated_at', 'tainted'])

    def test_node_create(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        node = self.node_create(pf['id'], name)
        self.assertEqual(node['name'], name)
        self.node_delete(node['id'])
        self.addCleanup(self.profile_delete, pf['id'])

    def test_node_update(self):
        old_name = self.name_generate()
        pf = self.profile_create(old_name)
        n1 = self.node_create(pf['id'], old_name)
        new_name = self.name_generate()
        pf_new = self.profile_create(new_name)
        role = 'master'
        tainted = 'True'
        cmd = ('cluster node update --name %s --role %s --profile %s '
               '--tainted %s %s'
               % (new_name, role, pf_new['id'], tainted, n1['id']))
        self.openstack(cmd)
        self.wait_for_status(n1['id'], 'ACTIVE', 'node', 120)
        raw_node = self.openstack('cluster node show %s' % n1['id'])
        node_data = self.show_to_dict(raw_node)
        self.assertEqual(node_data['name'], new_name)
        self.assertNotEqual(node_data['name'], old_name)
        self.assertEqual(node_data['role'], role)
        self.assertEqual(node_data['tainted'], tainted)
        self.assertEqual(node_data['profile_id'], pf_new['id'])
        self.node_delete(new_name)
        self.addCleanup(self.profile_delete, pf['id'])
        self.addCleanup(self.profile_delete, pf_new['id'])

    def test_node_detail(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        node = self.node_create(pf['id'], name)
        cmd = ('cluster node show --details %s' % name)
        raw_node = self.openstack(cmd)
        node_data = self.show_to_dict(raw_node)
        self.assertIn('details', node_data)
        self.assertIsNotNone(node_data['details'])
        self.node_delete(node['id'])
        self.addCleanup(self.profile_delete, pf['id'])

    # NOTE(Qiming): Since functional tests only focus on the client/server
    # interaction without invovling other OpenStack services, it is not
    # possible to mock a node failure and then test if the check logic works.
    # Such tests would be left to integration tests instead.
    def test_node_check(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        node = self.node_create(pf['id'], name)
        cmd = ('cluster node check %s' % node['id'])
        self.openstack(cmd)
        check_raw = self.openstack('cluster node show %s' % name)
        check_data = self.show_to_dict(check_raw)
        self.assertIn('Check', check_data['status_reason'])
        node_status = ['ACTIVE', 'ERROR']
        self.assertIn(check_data['status'], node_status)
        self.node_delete(node['id'])
        self.addCleanup(self.profile_delete, pf['id'])

    # NOTE(Qiming): A end-to-end test of the node recover operation needs to
    # be done with other OpenStack services involved, thus out of scope for
    # functional tests. Such tests would be left to integration tests instead.
    def test_node_recover(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        node = self.node_create(pf['id'], name)
        cmd = ('cluster node recover --check true %s' % node['id'])
        self.openstack(cmd)
        self.wait_for_status(name, 'ACTIVE', 'node', 120)
        recover_raw = self.openstack('cluster node show %s' % name)
        recover_data = self.show_to_dict(recover_raw)
        self.assertIn('Recover', recover_data['status_reason'])
        self.assertEqual('ACTIVE', recover_data['status'])
        self.node_delete(node['id'])
        self.addCleanup(self.profile_delete, pf['id'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc
from unittest import mock

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import profile_type as osc_profile_type


class TestProfileType(fakes.TestClusteringv1):
    def setUp(self):
        super(TestProfileType, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestProfileTypeList(TestProfileType):

    def setUp(self):
        super(TestProfileTypeList, self).setUp()
        self.cmd = osc_profile_type.ProfileTypeList(self.app, None)
        pt1 = mock.Mock(
            schema={'foo': 'bar'},
            support_status={
                "1.0": [{"status": "SUPPORTED", "since": "2016.10"}]
            }
        )
        pt1.name = "BBB"
        pt2 = mock.Mock(
            schema={'foo': 'bar'},
            support_status={
                "1.0": [{"status": "DEPRECATED", "since": "2016.01"}]
            }
        )
        pt2.name = "AAA"
        self.mock_client.profile_types = mock.Mock(return_value=[pt1, pt2])

    def test_profile_type_list(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        expected_rows = [
            ('AAA', '1.0', 'DEPRECATED since 2016.01'),
            ('BBB', '1.0', 'SUPPORTED since 2016.10')
        ]
        expected_columns = ['name', 'version', 'support_status']

        columns, rows = self.cmd.take_action(parsed_args)
        if len(columns) == 2:
            expected_columns = ['name', 'version']
            expected_rows = [
                ('CCC', '1.0')
            ]

        self.mock_client.profile_types.assert_called_with()
        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_rows, rows)


class TestProfileTypeShow(TestProfileType):

    def setUp(self):
        super(TestProfileTypeShow, self).setUp()
        self.cmd = osc_profile_type.ProfileTypeShow(self.app, None)
        fake_profile_type = mock.Mock(
            schema={'foo': 'bar'},
            support_status={
                "1.0": [{"status": "DEPRECATED", "since": "2016.01"}]
            }
        )
        fake_profile_type.name = "os.heat.stack-1.0"
        fake_profile_type.to_dict = mock.Mock(return_value={})
        self.mock_client.get_profile_type = mock.Mock(
            return_value=fake_profile_type)

    def test_profile_type_show(self):
        arglist = ['os.heat.stack-1.0']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_profile_type.assert_called_once_with(
            'os.heat.stack-1.0')

    def test_profile_type_show_not_found(self):
        arglist = ['os.heat.stack-1.1']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_profile_type.side_effect = (
            sdk_exc.ResourceNotFound())
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Profile Type not found: os.heat.stack-1.1',
                         str(error))


class TestProfileTypeOperations(TestProfileType):
    def setUp(self):
        super(TestProfileTypeOperations, self).setUp()
        self.cmd = osc_profile_type.ProfileTypeOperations(self.app, None)
        fake_profile_type_ops = mock.Mock(
            {
                'options': {
                    'abandon': {
                        'required': False,
                        'type': 'Map',
                        'description': 'Abandon a heat stack node.',
                        'updatable': False
                    }
                }
            }
        )
        self.mock_client.list_profile_type_operations = mock.Mock(
            return_value=fake_profile_type_ops)

    def test_profile_type_operations(self):
        arglist = ['os.heat.stack-1.0']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.list_profile_type_operations.assert_called_once_with(
            'os.heat.stack-1.0')

    def test_profile_type_operations_not_found(self):
        arglist = ['os.heat.stack-1.1']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.list_profile_type_operations.side_effect = (
            sdk_exc.ResourceNotFound())
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Profile Type not found: os.heat.stack-1.1',
                         str(error))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import event as osc_event


class TestEvent(fakes.TestClusteringv1):
    def setUp(self):
        super(TestEvent, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestEventList(TestEvent):

    columns = ['id', 'generated_at', 'obj_type', 'obj_id', 'obj_name',
               'action', 'status', 'level', 'cluster_id', 'meta_data']
    defaults = {
        'global_project': False,
        'marker': None,
        'limit': None,
        'sort': None,
    }

    def setUp(self):
        super(TestEventList, self).setUp()
        self.cmd = osc_event.ListEvent(self.app, None)
        fake_event = mock.Mock(
            action="CREATE",
            cluster_id=None,
            id="2d255b9c-8f36-41a2-a137-c0175ccc29c3",
            level="20",
            obj_id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            obj_name="node009",
            obj_type="NODE",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="CREATING",
            generated_at="2015-03-05T08:53:15",
            user_id="a21ded6060534d99840658a777c2af5a"
        )
        fake_event.to_dict = mock.Mock(return_value={})
        self.mock_client.events = mock.Mock(return_value=[fake_event])

    def test_event_list_defaults(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.events.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_event_list_full_id(self):
        arglist = ['--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.events.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_event_list_limit(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['limit'] = '3'
        arglist = ['--limit', '3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.events.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_event_list_sort(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:asc'
        arglist = ['--sort', 'name:asc']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.events.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_event_list_sort_invalid_key(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'bad_key'
        arglist = ['--sort', 'bad_key']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.events.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_event_list_sort_invalid_direction(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:bad_direction'
        arglist = ['--sort', 'name:bad_direction']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.events.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_event_list_filter(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['name'] = 'my_event'
        arglist = ['--filter', 'name=my_event']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.events.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_event_list_marker(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['marker'] = 'a9448bf6'
        arglist = ['--marker', 'a9448bf6']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.events.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)


class TestEventShow(TestEvent):

    def setUp(self):
        super(TestEventShow, self).setUp()
        self.cmd = osc_event.ShowEvent(self.app, None)
        fake_event = mock.Mock(
            action="CREATE",
            cluster_id=None,
            id="2d255b9c-8f36-41a2-a137-c0175ccc29c3",
            level="20",
            obj_id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            obj_name="node009",
            obj_type="NODE",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="CREATING",
            generated_at="2015-03-05T08:53:15",
            user_id="a21ded6060534d99840658a777c2af5a"
        )
        fake_event.to_dict = mock.Mock(return_value={})
        self.mock_client.get_event = mock.Mock(return_value=fake_event)

    def test_event_show(self):
        arglist = ['my_event']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.get_event.assert_called_with('my_event')

    def test_event_show_not_found(self):
        arglist = ['my_event']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_event.side_effect = sdk_exc.ResourceNotFound()

        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)

        self.assertEqual('Event not found: my_event', str(error))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import build_info as osc_build_info


class TestBuildInfo(fakes.TestClusteringv1):

    def setUp(self):
        super(TestBuildInfo, self).setUp()
        self.cmd = osc_build_info.BuildInfo(self.app, None)
        self.mock_client = self.app.client_manager.clustering
        fake_bi = mock.Mock(
            api={"revision": "1.0"},
            engine={"revision": "1.0"}
        )
        self.mock_client.get_build_info = mock.Mock(return_value=fake_bi)

    def test_build_info(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_build_info.assert_called_with()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class PolicyTypeTest(base.OpenStackClientTestBase):
    """Test for policy types"""

    def test_policy_type_list(self):
        result = self.openstack('cluster policy type list')
        policy_type = self.parser.listing(result)
        columns = ['name', 'version']
        if any('support_status' in i for i in policy_type):
            columns.append('support_status')
        self.assertTableStruct(policy_type, columns)

    def test_policy_type_show(self):
        params = ['senlin.policy.affinity-1.0',
                  'senlin.policy.batch-1.0',
                  'senlin.policy.deletion-1.0',
                  'senlin.policy.health-1.0',
                  'senlin.policy.loadbalance-1.1',
                  'senlin.policy.region_placement-1.0',
                  'senlin.policy.zone_placement-1.0']
        for param in params:
            cmd = 'cluster policy type show %s' % param
            self.openstack(cmd)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 node action implementations"""

import logging
import sys

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils
from oslo_utils import strutils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ListNode(command.Lister):
    """List nodes."""

    log = logging.getLogger(__name__ + ".ListNode")

    def get_parser(self, prog_name):
        parser = super(ListNode, self).get_parser(prog_name)
        parser.add_argument(
            '--cluster',
            metavar='<cluster>',
            help=_('ID or name of cluster from which nodes are to be listed')
        )
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned nodes. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter"
                   " keys are: ['status','name']"),
            action='append'
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc). The valid "
                   "sort keys are:['index', 'name', 'status', 'init_at', "
                   "'created_at', 'updated_at']")
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of nodes returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return nodes that appear after the given node ID')
        )
        parser.add_argument(
            '--global-project',
            default=False, action="store_true",
            help=_('Indicate that this node list should include nodes from '
                   'all projects. This option is subject to access policy '
                   'checking. Default is False')
        )
        parser.add_argument(
            '--full-id',
            default=False, action="store_true",
            help=_('Print full IDs in list')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering

        columns = ['id', 'name', 'index', 'status', 'cluster_id',
                   'physical_id', 'profile_name', 'created_at', 'updated_at',
                   'tainted']
        queries = {
            'cluster_id': parsed_args.cluster,
            'sort': parsed_args.sort,
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
            'global_project': parsed_args.global_project,
        }

        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        nodes = senlin_client.nodes(**queries)
        if parsed_args.global_project:
            columns.append('project_id')
        if not parsed_args.full_id:
            formatters = {
                'id': lambda x: x[:8],
                'cluster_id': lambda x: x[:8] if x else '',
                'physical_id': lambda x: x[:8] if x else ''
            }
            if 'project_id' in columns:
                formatters['project_id'] = lambda x: x[:8]
        else:
            formatters = {}

        return (
            columns,
            (utils.get_item_properties(n, columns, formatters=formatters)
             for n in nodes)
        )


class ShowNode(command.ShowOne):
    """Show detailed info about the specified node."""

    log = logging.getLogger(__name__ + ".ShowNode")

    def get_parser(self, prog_name):
        parser = super(ShowNode, self).get_parser(prog_name)
        parser.add_argument(
            '--details',
            default=False,
            action="store_true",
            help=_('Include physical object details')
        )
        parser.add_argument(
            'node',
            metavar='<node>',
            help=_('Name or ID of the node to show the details for')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        return _show_node(senlin_client, parsed_args.node, parsed_args.details)


def _show_node(senlin_client, node_id, show_details=False):
    """Show detailed info about the specified node."""

    try:
        node = senlin_client.get_node(node_id, details=show_details)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('Node not found: %s') % node_id)

    formatters = {
        'metadata': senlin_utils.json_formatter,
        'data': senlin_utils.json_formatter,
        'dependents': senlin_utils.json_formatter,
    }
    data = node.to_dict()
    if show_details and data['details']:
        formatters['details'] = senlin_utils.nested_dict_formatter(
            list(data['details'].keys()), ['property', 'value'])

    columns = sorted(data.keys())
    return columns, utils.get_dict_properties(data, columns,
                                              formatters=formatters)


class CreateNode(command.ShowOne):
    """Create the node."""

    log = logging.getLogger(__name__ + ".CreateNode")

    def get_parser(self, prog_name):
        parser = super(CreateNode, self).get_parser(prog_name)
        parser.add_argument(
            '--cluster',
            metavar='<cluster>',
            help=_('Cluster Id or Name for this node')
        )
        parser.add_argument(
            '--role',
            metavar='<role>',
            help=_('Role for this node in the specific cluster')
        )
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_('Metadata values to be attached to the node. '
                   'This can be specified multiple times, or once with '
                   'key-value pairs separated by a semicolon'),
            action='append'
        )
        parser.add_argument(
            '--profile',
            metavar='<profile>',
            required=True,
            help=_('Profile Id or Name used for this node')
        )
        parser.add_argument(
            'name',
            metavar='<node-name>',
            help=_('Name of the node to create')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        attrs = {
            'name': parsed_args.name,
            'cluster_id': parsed_args.cluster,
            'profile_id': parsed_args.profile,
            'role': parsed_args.role,
            'metadata': senlin_utils.format_parameters(parsed_args.metadata),
        }

        node = senlin_client.create_node(**attrs)
        return _show_node(senlin_client, node.id)


class UpdateNode(command.ShowOne):
    """Update the node."""

    log = logging.getLogger(__name__ + ".UpdateNode")

    def get_parser(self, prog_name):
        parser = super(UpdateNode, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New name for the node')
        )
        parser.add_argument(
            '--profile',
            metavar='<profile>',
            help=_('ID or name of new profile to use')
        )
        parser.add_argument(
            '--role',
            metavar='<role>',
            help=_('Role for this node in the specific cluster')
        )
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Metadata values to be attached to the node. "
                   "This can be specified multiple times, or once with "
                   "key-value pairs separated by a semicolon. Use '{}' "
                   "can clean metadata "),
            action='append'
        )
        parser.add_argument(
            '--tainted',
            metavar='<boolean>',
            help=_("Whether the node should be marked as tainted. "
                   "If true, this node will be selected first for the next"
                   "cluster scale-in operation.")

        )
        parser.add_argument(
            'node',
            metavar='<node>',
            help=_('Name or ID of node to update')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering

        # Find the node first, we need its UUID
        node = senlin_client.find_node(parsed_args.node)
        if node is None:
            raise exc.CommandError(_('Node not found: %s') % parsed_args.node)

        attrs = {
            'name': parsed_args.name,
            'role': parsed_args.role,
            'profile_id': parsed_args.profile,
            'metadata': senlin_utils.format_parameters(parsed_args.metadata),
        }

        if parsed_args.tainted is not None:
            attrs['tainted'] = strutils.bool_from_string(
                parsed_args.tainted,
                strict=True,
            )

        senlin_client.update_node(node.id, **attrs)
        return _show_node(senlin_client, node.id)


class DeleteNode(command.Command):
    """Delete the node(s)."""

    log = logging.getLogger(__name__ + ".DeleteNode")

    def get_parser(self, prog_name):
        parser = super(DeleteNode, self).get_parser(prog_name)
        parser.add_argument(
            'node',
            metavar='<node>',
            nargs='+',
            help=_('Name or ID of node(s) to delete.')
        )
        parser.add_argument(
            '--force-delete',
            action='store_true',
            help=_('Force to delete node(s).')
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help=_('Skip yes/no prompt (assume yes).')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        try:
            if not parsed_args.force and sys.stdin.isatty():
                sys.stdout.write(
                    _("Are you sure you want to delete this node(s)"
                      " [y/N]?"))
                sys.stdout.flush()
                prompt_response = sys.stdin.readline().lower()
                if not prompt_response.startswith('y'):
                    return
        except KeyboardInterrupt:  # Ctrl-c
            self.log.info('Ctrl-c detected.')
            return
        except EOFError:  # Ctrl-d
            self.log.info('Ctrl-d detected')
            return

        result = {}
        for nid in parsed_args.node:
            try:
                node_delete_action = senlin_client.delete_node(
                    nid, False, parsed_args.force_delete)
                result[nid] = ('OK', node_delete_action['id'])
            except Exception as ex:
                result[nid] = ('ERROR', str(ex))

        for rid, res in result.items():
            senlin_utils.print_action_result(rid, res)


class CheckNode(command.Command):
    """Check the node(s)."""
    log = logging.getLogger(__name__ + ".CheckNode")

    def get_parser(self, prog_name):
        parser = super(CheckNode, self).get_parser(prog_name)
        parser.add_argument(
            'node',
            metavar='<node>',
            nargs='+',
            help=_('ID or name of node(s) to check.')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        for nid in parsed_args.node:
            try:
                resp = senlin_client.check_node(nid)
            except sdk_exc.ResourceNotFound:
                raise exc.CommandError(_('Node not found: %s') % nid)
            if 'action' in resp:
                print('Node check request on node %(nid)s is accepted by '
                      'action %(action)s.'
                      % {'nid': nid, 'action': resp['action']})
            else:
                print('Request error: %s' % resp)


class RecoverNode(command.Command):
    """Recover the node(s)."""
    log = logging.getLogger(__name__ + ".RecoverNode")

    def get_parser(self, prog_name):
        parser = super(RecoverNode, self).get_parser(prog_name)
        parser.add_argument(
            '--check',
            metavar='<boolean>',
            default=False,
            help=_('Whether the node(s) should check physical resource status '
                   'before doing node recover. Default is false')
        )
        parser.add_argument(
            'node',
            metavar='<node>',
            nargs='+',
            help=_('ID or name of node(s) to recover.')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        params = {
            'check': strutils.bool_from_string(parsed_args.check, strict=True)
        }

        for nid in parsed_args.node:
            try:
                resp = senlin_client.recover_node(nid, **params)
            except sdk_exc.ResourceNotFound:
                raise exc.CommandError(_('Node not found: %s') % nid)
            if 'action' in resp:
                print('Node recover request on node %(nid)s is accepted by '
                      'action %(action)s.'
                      % {'nid': nid, 'action': resp['action']})
            else:
                print('Request error: %s' % resp)


class AdoptNode(command.ShowOne):
    """Adopt (or preview) the node."""

    log = logging.getLogger(__name__ + ".AdoptNode")

    def get_parser(self, prog_name):
        parser = super(AdoptNode, self).get_parser(prog_name)
        parser.add_argument(
            '--identity',
            metavar='<identity>',
            required=True,
            help=_('Physical resource id.'))
        parser.add_argument(
            '--type',
            metavar='<type>',
            required=True,
            help=_('The name of the profile type.')
        )
        parser.add_argument(
            '--role',
            metavar='<role>',
            help=_('Role for this node in the specific cluster.')
        )
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_('Metadata values to be attached to the node. '
                   'This can be specified multiple times, or once with '
                   'key-value pairs separated by a semicolon.'),
            action='append'
        )
        parser.add_argument(
            '--name',
            metavar='<node-name>',
            help=_('Name of the node to adopt.')
        )
        parser.add_argument(
            '--overrides',
            metavar='<json>',
            help=_('JSON formatted specification for overriding this node '
                   'properties.')
        )
        parser.add_argument(
            '--preview',
            default=False,
            help=_('Whether preview the node adopt request. If set, '
                   'only previewing this node and do not adopt.'),
            action='store_true',
        )
        parser.add_argument(
            '--snapshot',
            default=False,
            help=_('Whether a shapshot of the existing physical object '
                   'should be created before the object is adopted as '
                   'a node.'),
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering

        preview = True if parsed_args.preview else False
        attrs = {
            'identity': parsed_args.identity,
            'overrides': senlin_utils.format_json_parameter(
                parsed_args.overrides),
            'snapshot': parsed_args.snapshot,
            'type': parsed_args.type
        }

        if not preview:
            attrs.update({
                'name': parsed_args.name,
                'role': parsed_args.role,
                'metadata': senlin_utils.format_parameters(
                    parsed_args.metadata),
            })

        node = senlin_client.adopt_node(preview, **attrs)

        if not preview:
            return _show_node(senlin_client, node.id)
        else:
            formatters = {}
            formatters['node_preview'] = senlin_utils.nested_dict_formatter(
                ['type', 'version', 'properties'],
                ['property', 'value'])
            data = node['node_profile']
            columns = sorted(data.keys())
            return columns, utils.get_dict_properties(data, columns,
                                                      formatters=formatters)


class NodeOp(command.Lister):
    """Perform an operation on a node."""
    log = logging.getLogger(__name__ + ".NodeOp")

    def get_parser(self, prog_name):
        parser = super(NodeOp, self).get_parser(prog_name)
        parser.add_argument(
            '--operation',
            metavar='<operation>',
            required=True,
            help=_('Operation to be performed on the node.')
        )
        parser.add_argument(
            '--params',
            metavar='<key1=value1;key2=value2...>',
            help=_("Parameters to for the specified operation. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon."),
            action="append"
        )
        parser.add_argument(
            'node',
            metavar='<node>',
            help=_('ID or name the node operate on.')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        nid = parsed_args.node
        if parsed_args.params:
            params = senlin_utils.format_parameters(parsed_args.params)
        else:
            params = {}

        try:
            resp = senlin_client.perform_operation_on_node(
                nid, parsed_args.operation, **params)
        except sdk_exc.ResourceNotFound:
            raise exc.CommandError(_('Node not found: %s') % nid)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
        else:
            print('Request error: %s' % resp)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import io
import subprocess
from unittest import mock

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import cluster as osc_cluster


class TestCluster(fakes.TestClusteringv1):
    def setUp(self):
        super(TestCluster, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestClusterList(TestCluster):

    columns = ['id', 'name', 'status', 'created_at', 'updated_at']
    defaults = {
        'global_project': False,
        'marker': None,
        'limit': None,
        'sort': None,
    }

    def setUp(self):
        super(TestClusterList, self).setUp()
        self.cmd = osc_cluster.ListCluster(self.app, None)
        fake_cluster = mock.Mock(
            created_at="2015-02-10T14:26:14",
            data={},
            desired_capacity=4,
            domain_id=None,
            id="7d85f602-a948-4a30-afd4-e84f47471c15",
            init_time="2015-02-10T14:26:11",
            max_size=-1,
            metadata={},
            min_size=0,
            node_ids=[
                "b07c57c8-7ab2-47bf-bdf8-e894c0c601b9",
                "ecc23d3e-bb68-48f8-8260-c9cf6bcb6e61",
                "da1e9c87-e584-4626-a120-022da5062dac"
            ],
            policies=[],
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="ACTIVE",
            status_reason="Cluster scale-in succeeded",
            timeout=3600,
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_cluster.name = "cluster1"
        fake_cluster.to_dict = mock.Mock(return_value={})

        self.mock_client.clusters = mock.Mock(return_value=[fake_cluster])

    def test_cluster_list_defaults(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.clusters.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_cluster_list_full_id(self):
        arglist = ['--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.clusters.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_cluster_list_limit(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['limit'] = '3'
        arglist = ['--limit', '3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.clusters.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_cluster_list_sort(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:asc'
        arglist = ['--sort', 'name:asc']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.clusters.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_cluster_list_sort_invalid_key(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'bad_key'
        arglist = ['--sort', 'bad_key']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.clusters.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_cluster_list_sort_invalid_direction(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:bad_direction'
        arglist = ['--sort', 'name:bad_direction']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.clusters.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_cluster_list_filter(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['name'] = 'my_cluster'
        arglist = ['--filter', 'name=my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.clusters.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_cluster_list_marker(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['marker'] = 'a9448bf6'
        arglist = ['--marker', 'a9448bf6']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.clusters.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)


class TestClusterShow(TestCluster):

    def setUp(self):
        super(TestClusterShow, self).setUp()
        self.cmd = osc_cluster.ShowCluster(self.app, None)
        fake_cluster = mock.Mock(
            config={},
            created_at="2015-02-11T15:13:20",
            data={},
            desired_capacity=0,
            domain_id=None,
            id="7d85f602-a948-4a30-afd4-e84f47471c15",
            init_time="2015-02-10T14:26:11",
            max_size=-1,
            metadata={},
            min_size=0,
            node_ids=[],
            policies=[],
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="ACTIVE",
            status_reason="Cluster scale-in succeeded",
            timeout=3600,
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_cluster.name = "my_cluster"
        fake_cluster.to_dict = mock.Mock(return_value={})
        self.mock_client.get_cluster = mock.Mock(return_value=fake_cluster)

    def test_cluster_show(self):
        arglist = ['my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_cluster.assert_called_with('my_cluster')

    def test_cluster_show_not_found(self):
        arglist = ['my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_cluster.side_effect = sdk_exc.ResourceNotFound()
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Cluster not found: my_cluster', str(error))


class TestClusterCreate(TestCluster):

    defaults = {
        "config": {},
        "desired_capacity": 0,
        "max_size": -1,
        "metadata": {},
        "min_size": 0,
        "name": "test_cluster",
        "profile_id": "mystack",
        "timeout": None
    }

    def setUp(self):
        super(TestClusterCreate, self).setUp()
        self.cmd = osc_cluster.CreateCluster(self.app, None)
        fake_cluster = mock.Mock(
            config={},
            created_at="2015-02-11T15:13:20",
            data={},
            desired_capacity=0,
            domain_id=None,
            id="7d85f602-a948-4a30-afd4-e84f47471c15",
            init_time="2015-02-10T14:26:11",
            max_size=-1,
            metadata={},
            min_size=0,
            node_ids=[],
            policies=[],
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="ACTIVE",
            status_reason="Cluster scale-in succeeded",
            timeout=3600,
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_cluster.name = "my_cluster"
        fake_cluster.to_dict = mock.Mock(return_value={})
        self.mock_client.create_cluster = mock.Mock(return_value=fake_cluster)
        self.mock_client.get_cluster = mock.Mock(return_value=fake_cluster)

    def test_cluster_create_defaults(self):
        arglist = ['test_cluster', '--profile', 'mystack']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_cluster.assert_called_with(**self.defaults)

    def test_cluster_create_with_metadata(self):
        arglist = ['test_cluster', '--profile', 'mystack',
                   '--metadata', 'key1=value1;key2=value2']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['metadata'] = {'key1': 'value1', 'key2': 'value2'}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_cluster.assert_called_with(**kwargs)

    def test_cluster_create_with_config(self):
        arglist = ['test_cluster', '--profile', 'mystack',
                   '--config', 'key1=value1;key2=value2']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['config'] = {'key1': 'value1', 'key2': 'value2'}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_cluster.assert_called_with(**kwargs)

    def test_cluster_create_with_size(self):
        arglist = ['test_cluster', '--profile', 'mystack',
                   '--min-size', '1', '--max-size', '10',
                   '--desired-capacity', '2']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['min_size'] = '1'
        kwargs['max_size'] = '10'
        kwargs['desired_capacity'] = '2'
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_cluster.assert_called_with(**kwargs)


class TestClusterUpdate(TestCluster):

    defaults = {
        "metadata": {
            "nk1": "nv1",
            "nk2": "nv2",
        },
        "name": 'new_cluster',
        "profile_id": 'new_profile',
        "profile_only": False,
        "timeout": "30"
    }

    def setUp(self):
        super(TestClusterUpdate, self).setUp()
        self.cmd = osc_cluster.UpdateCluster(self.app, None)
        self.fake_cluster = mock.Mock(
            created_at="2015-02-11T15:13:20",
            data={},
            desired_capacity=0,
            domain_id=None,
            id="7d85f602-a948-4a30-afd4-e84f47471c15",
            init_time="2015-02-10T14:26:11",
            max_size=-1,
            metadata={},
            min_size=0,
            node_ids=[],
            policies=[],
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="ACTIVE",
            status_reason="Cluster scale-in succeeded",
            timeout=3600,
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        self.fake_cluster.name = "my_cluster"
        self.fake_cluster.to_dict = mock.Mock(return_value={})
        self.mock_client.update_cluster = mock.Mock(
            return_value=self.fake_cluster)
        self.mock_client.get_cluster = mock.Mock(
            return_value=self.fake_cluster)
        self.mock_client.find_cluster = mock.Mock(
            return_value=self.fake_cluster)

    def test_cluster_update_defaults(self):
        arglist = ['--name', 'new_cluster', '--metadata', 'nk1=nv1;nk2=nv2',
                   '--profile', 'new_profile', '--timeout', '30', '45edadcb']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.update_cluster.assert_called_with(
            self.fake_cluster, **self.defaults)

    def test_cluster_update_not_found(self):
        arglist = ['--name', 'new_cluster', '--metadata', 'nk1=nv1;nk2=nv2',
                   '--profile', 'new_profile', 'c6b8b252']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.find_cluster.return_value = None
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Cluster not found: c6b8b252', str(error))


class TestClusterDelete(TestCluster):
    def setUp(self):
        super(TestClusterDelete, self).setUp()
        self.cmd = osc_cluster.DeleteCluster(self.app, None)
        mock_cluster = mock.Mock(location='abc/fake_action_id')
        self.mock_client.delete_cluster = mock.Mock(return_value=mock_cluster)

    def test_cluster_delete(self):
        arglist = ['cluster1', 'cluster2', 'cluster3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_cluster.assert_has_calls(
            [mock.call('cluster1', False, False),
             mock.call('cluster2', False, False),
             mock.call('cluster3', False, False)]
        )

    def test_cluster_delete_force(self):
        arglist = ['cluster1', 'cluster2', 'cluster3', '--force']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_cluster.assert_has_calls(
            [mock.call('cluster1', False, False),
             mock.call('cluster2', False, False),
             mock.call('cluster3', False, False)]
        )

    def test_cluster_delete_force_delete(self):
        arglist = ['cluster1', 'cluster2', 'cluster3', '--force-delete']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_cluster.assert_has_calls(
            [mock.call('cluster1', False, True),
             mock.call('cluster2', False, True),
             mock.call('cluster3', False, True)]
        )

    def test_cluster_delete_not_found(self):
        arglist = ['my_cluster']
        self.mock_client.delete_cluster.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.delete_cluster.assert_has_calls(
            [mock.call('my_cluster', False, False)]
        )

    def test_cluster_delete_one_found_one_not_found(self):
        arglist = ['cluster1', 'cluster2']
        self.mock_client.delete_cluster.side_effect = (
            [None, sdk_exc.ResourceNotFound]
        )
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.delete_cluster.assert_has_calls(
            [mock.call('cluster1', False, False),
             mock.call('cluster2', False, False)]
        )

    @mock.patch('sys.stdin', spec=io.StringIO)
    def test_cluster_delete_prompt_yes(self, mock_stdin):
        arglist = ['my_cluster']
        mock_stdin.isatty.return_value = True
        mock_stdin.readline.return_value = 'y'
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        mock_stdin.readline.assert_called_with()
        self.mock_client.delete_cluster.assert_called_with(
            'my_cluster', False, False)

    @mock.patch('sys.stdin', spec=io.StringIO)
    def test_cluster_delete_prompt_no(self, mock_stdin):
        arglist = ['my_cluster']
        mock_stdin.isatty.return_value = True
        mock_stdin.readline.return_value = 'n'
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        mock_stdin.readline.assert_called_with()
        self.mock_client.delete_cluster.assert_not_called()


class TestClusterResize(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}
    defaults = {
        "min_step": None,
        "adjustment_type": "EXACT_CAPACITY",
        "number": 2,
        "min_size": 1,
        "strict": True,
        "max_size": 20}

    def setUp(self):
        super(TestClusterResize, self).setUp()
        self.cmd = osc_cluster.ResizeCluster(self.app, None)
        self.mock_client.resize_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_resize_no_params(self):
        arglist = ['my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual(("At least one parameter of 'capacity', "
                          "'adjustment', 'percentage', 'min_size' and "
                          "'max_size' should be specified."), str(error))

    def test_cluster_resize_multi_params(self):
        arglist = ['--capacity', '2', '--percentage', '50.0', '--adjustment',
                   '1', '--min-size', '1', '--max-size', '20', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual("Only one of 'capacity', 'adjustment' "
                         "and 'percentage' can be specified.", str(error))

    def test_cluster_resize_capacity(self):
        arglist = ['--capacity', '2', '--min-size', '1', '--max-size', '20',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.resize_cluster.assert_called_with('my_cluster',
                                                           **self.defaults)

    def test_cluster_resize_invalid_capacity(self):
        arglist = ['--capacity', '-1', '--min-size', '1', '--max-size', '20',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Cluster capacity must be larger than or equal to'
                         ' zero.', str(error))

    def test_cluster_resize_adjustment(self):
        arglist = ['--adjustment', '1', '--min-size', '1', '--max-size', '20',
                   'my_cluster', '--strict']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['adjustment_type'] = 'CHANGE_IN_CAPACITY'
        kwargs['number'] = 1
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.resize_cluster.assert_called_with('my_cluster',
                                                           **kwargs)

    def test_cluster_resize_invalid_adjustment(self):
        arglist = ['--adjustment', '0', '--min-size', '1', '--max-size', '20',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Adjustment cannot be zero.', str(error))

    def test_cluster_resize_percentage(self):
        arglist = ['--percentage', '50.0', '--min-size', '1', '--max-size',
                   '20', 'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        kwargs = copy.deepcopy(self.defaults)
        kwargs['adjustment_type'] = 'CHANGE_IN_PERCENTAGE'
        kwargs['number'] = 50.0
        self.cmd.take_action(parsed_args)
        self.mock_client.resize_cluster.assert_called_with('my_cluster',
                                                           **kwargs)

    def test_cluster_resize_invalid_percentage(self):
        arglist = ['--percentage', '0', '--min-size', '1', '--max-size', '20',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Percentage cannot be zero.', str(error))

    def test_cluster_resize_invalid_min_step_capacity(self):
        arglist = ['--capacity', '2', '--min-size', '1', '--max-size', '20',
                   'my_cluster', '--strict', '--min-step', '1']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Min step is only used with percentage.', str(error))

    def test_cluster_resize_invalid_min_size_capacity(self):
        arglist = ['--capacity', '2', '--min-size', '-1', '--max-size', '20',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Min size cannot be less than zero.', str(error))

    def test_cluster_resize_invalid_max_size_capacity(self):
        arglist = ['--capacity', '2', '--min-size', '5', '--max-size', '3',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Min size cannot be larger than max size.',
                         str(error))

    def test_cluster_resize_min_size_larger_than_capacity(self):
        arglist = ['--capacity', '3', '--min-size', '5', '--max-size', '10',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Min size cannot be larger than the specified '
                         'capacity', str(error))

    def test_cluster_resize_invalid_max_size_less_capacity(self):
        arglist = ['--capacity', '15', '--min-size', '5', '--max-size', '10',
                   'my_cluster', '--strict']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Max size cannot be less than the specified '
                         'capacity.', str(error))


class TestClusterScaleIn(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterScaleIn, self).setUp()
        self.cmd = osc_cluster.ScaleInCluster(self.app, None)
        self.mock_client.scale_in_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_scale_in(self):
        arglist = ['--count', '2', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.scale_in_cluster.assert_called_with('my_cluster',
                                                             '2')


class TestClusterScaleOut(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterScaleOut, self).setUp()
        self.cmd = osc_cluster.ScaleOutCluster(self.app, None)
        self.mock_client.scale_out_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_scale_out(self):
        arglist = ['--count', '2', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.scale_out_cluster.assert_called_with('my_cluster',
                                                              '2')


class TestClusterPolicyAttach(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterPolicyAttach, self).setUp()
        self.cmd = osc_cluster.ClusterPolicyAttach(self.app, None)
        self.mock_client.attach_policy_to_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_policy_attach(self):
        arglist = ['--policy', 'my_policy', '--enabled', 'True', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.attach_policy_to_cluster.assert_called_with(
            'my_cluster',
            'my_policy',
            enabled=True)


class TestClusterPolicyDetach(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterPolicyDetach, self).setUp()
        self.cmd = osc_cluster.ClusterPolicyDetach(self.app, None)
        self.mock_client.detach_policy_from_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_policy_detach(self):
        arglist = ['--policy', 'my_policy', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.detach_policy_from_cluster.assert_called_with(
            'my_cluster',
            'my_policy')


class TestClusterNodeList(TestCluster):
    columns = ['id', 'name', 'index', 'status', 'physical_id', 'created_at']
    args = {
        'cluster_id': 'my_cluster',
        'marker': 'a9448bf6',
        'limit': '3',
        'sort': None,
    }

    def setUp(self):
        super(TestClusterNodeList, self).setUp()
        self.cmd = osc_cluster.ClusterNodeList(self.app, None)
        fake_node = mock.Mock(
            cluster_id="",
            created_at="2015-02-11T15:13:20",
            data={},
            details={},
            domain_id=None,
            id="7d85f602-a948-4a30-afd4-e84f47471c15",
            index=-1,
            init_at="2015-02-10T14:26:11",
            metadata={},
            phyiscal_id="cc028275-d078-4729-bf3e-154b7359814b",
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            status="ACTIVE",
            status_reason="Creation succeeded",
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_node.name = "node001"
        fake_node.to_dict = mock.Mock(return_value={})
        self.mock_client.nodes = mock.Mock(return_value=[fake_node])

    def test_cluster_node_list(self):
        arglist = ['--limit', '3', '--marker', 'a9448bf6', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**self.args)
        self.assertEqual(self.columns, columns)

    def test_cluster_node_list_full_id(self):
        arglist = ['--limit', '3', '--marker', 'a9448bf6', 'my_cluster',
                   '--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**self.args)
        self.assertEqual(self.columns, columns)

    def test_cluster_node_list_filter(self):
        kwargs = copy.deepcopy(self.args)
        kwargs['name'] = 'my_node'
        arglist = ['--limit', '3', '--marker', 'a9448bf6', 'my_cluster',
                   '--filter', 'name=my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_cluster_node_list_sort(self):
        kwargs = copy.deepcopy(self.args)
        kwargs['name'] = 'my_node'
        kwargs['sort'] = 'name:asc'
        arglist = ['--limit', '3', '--marker', 'a9448bf6', 'my_cluster',
                   '--filter', 'name=my_node', '--sort', 'name:asc']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)


class TestClusterNodeAdd(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterNodeAdd, self).setUp()
        self.cmd = osc_cluster.ClusterNodeAdd(self.app, None)
        self.mock_client.add_nodes_to_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_node_add(self):
        arglist = ['--nodes', 'node1', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.add_nodes_to_cluster.assert_called_with(
            'my_cluster',
            ['node1'])

    def test_cluster_node_add_multi(self):
        arglist = ['--nodes', 'node1,node2', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.add_nodes_to_cluster.assert_called_with(
            'my_cluster',
            ['node1', 'node2'])


class TestClusterNodeDel(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterNodeDel, self).setUp()
        self.cmd = osc_cluster.ClusterNodeDel(self.app, None)
        self.mock_client.remove_nodes_from_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_node_delete(self):
        arglist = ['-d', 'True', '--nodes', 'node1', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.remove_nodes_from_cluster.assert_called_with(
            'my_cluster',
            ['node1'],
            destroy_after_deletion=True)

    def test_cluster_node_delete_without_destroy(self):
        arglist = ['-d', 'False', '--nodes', 'node1', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.remove_nodes_from_cluster.assert_called_with(
            'my_cluster',
            ['node1'],
            destroy_after_deletion=False)

    def test_cluster_node_delete_multi(self):
        arglist = ['--nodes', 'node1,node2', 'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.remove_nodes_from_cluster.assert_called_with(
            'my_cluster',
            ['node1', 'node2'],
            destroy_after_deletion=False)


class TestClusterCheck(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterCheck, self).setUp()
        self.cmd = osc_cluster.CheckCluster(self.app, None)
        self.mock_client.check_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_check(self):
        arglist = ['cluster1', 'cluster2', 'cluster3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.check_cluster.assert_has_calls(
            [mock.call('cluster1'), mock.call('cluster2'),
             mock.call('cluster3')]
        )

    def test_cluster_check_not_found(self):
        arglist = ['cluster1']
        self.mock_client.check_cluster.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Cluster not found: cluster1', str(error))


class TestClusterRecover(TestCluster):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestClusterRecover, self).setUp()
        self.cmd = osc_cluster.RecoverCluster(self.app, None)
        self.mock_client.recover_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_recover(self):
        arglist = ['cluster1', 'cluster2', 'cluster3', '--check', 'false']
        kwargs = {'check': False}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.recover_cluster.assert_has_calls(
            [mock.call('cluster1', **kwargs), mock.call('cluster2', **kwargs),
             mock.call('cluster3', **kwargs)]
        )

    def test_cluster_recover_not_found(self):
        arglist = ['cluster1']
        self.mock_client.recover_cluster.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Cluster not found: cluster1', str(error))


class TestClusterOp(TestCluster):

    response = {"action": "a3c6d04c-3fca-4e4a-b0b3-c0522ef711f1"}

    def setUp(self):
        super(TestClusterOp, self).setUp()
        self.cmd = osc_cluster.ClusterOp(self.app, None)
        self.mock_client.perform_operation_on_cluster = mock.Mock(
            return_value=self.response)

    def test_cluster_op(self):
        arglist = ['--operation', 'dance', '--params', 'style=tango',
                   'my_cluster']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.perform_operation_on_cluster.assert_called_once_with(
            'my_cluster',
            'dance',
            style='tango')

    def test_cluster_op_not_found(self):
        arglist = ['--operation', 'dance', 'cluster1']
        ex = sdk_exc.ResourceNotFound
        self.mock_client.perform_operation_on_cluster.side_effect = ex
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Cluster not found: cluster1', str(error))


class TestClusterCollect(TestCluster):
    response = [
        {
            "node_id": "8bb476c3-0f4c-44ee-9f64-c7b0260814de",
            "attr_value": "value 1",
        },
        {
            "node_id": "7d85f602-a948-4a30-afd4-e84f47471c15",
            "attr_value": "value 2",
        }
    ]

    def setUp(self):
        super(TestClusterCollect, self).setUp()
        self.cmd = osc_cluster.ClusterCollect(self.app, None)
        self.mock_client.collect_cluster_attrs = mock.Mock(
            return_value=self.response)

    def test_cluster_collect(self):
        arglist = ['--path', 'path.to.attr', 'cluster1']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.collect_cluster_attrs.assert_called_once_with(
            'cluster1', 'path.to.attr')
        self.assertEqual(['node_id', 'attr_value'], columns)

    def test_cluster_collect_with_full_id(self):
        arglist = ['--path', 'path.to.attr', '--full-id', 'cluster1']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.collect_cluster_attrs.assert_called_once_with(
            'cluster1', 'path.to.attr')
        self.assertEqual(['node_id', 'attr_value'], columns)


class TestClusterRun(TestCluster):
    attrs = [
        mock.Mock(node_id="NODE_ID1",
                  attr_value={"addresses": 'ADDRESS CONTENT 1'}),
        mock.Mock(node_id="NODE_ID2",
                  attr_value={"addresses": 'ADDRESS CONTENT 2'})
    ]

    def setUp(self):
        super(TestClusterRun, self).setUp()
        self.cmd = osc_cluster.ClusterRun(self.app, None)
        self.mock_client.collect_cluster_attrs = mock.Mock(
            return_value=self.attrs)

    @mock.patch('subprocess.Popen')
    def test__run_script(self, mock_proc):
        x_proc = mock.Mock(returncode=0)
        x_stdout = 'OUTPUT'
        x_stderr = 'ERROR'
        x_proc.communicate.return_value = (x_stdout, x_stderr)
        mock_proc.return_value = x_proc

        addr = {
            'private': [
                {
                    'OS-EXT-IPS:type': 'floating',
                    'version': 4,
                    'addr': '1.2.3.4',
                }
            ]
        }
        output = {}

        self.cmd._run_script('NODE_ID', addr, 'private', 'floating', 22,
                             'john', False, 'identity_path', 'echo foo',
                             '-f bar',
                             output=output)
        mock_proc.assert_called_once_with(
            ['ssh', '-4', '-p22', '-i identity_path', '-f bar', 'john@1.2.3.4',
             'echo foo'],
            stdout=subprocess.PIPE)
        self.assertEqual(
            {'status': 'SUCCEEDED (0)', 'output': 'OUTPUT', 'error': 'ERROR'},
            output)

    def test__run_script_network_not_found(self):
        addr = {'foo': 'bar'}
        output = {}

        self.cmd._run_script('NODE_ID', addr, 'private', 'floating', 22,
                             'john', False, 'identity_path', 'echo foo',
                             '-f bar',
                             output=output)
        self.assertEqual(
            {'status': 'FAILED',
             'error': "Node 'NODE_ID' is not attached to network 'private'."
             },
            output)

    def test__run_script_more_than_one_network(self):
        addr = {'foo': 'bar', 'koo': 'tar'}
        output = {}

        self.cmd._run_script('NODE_ID', addr, '', 'floating', 22, 'john',
                             False, 'identity_path', 'echo foo', '-f bar',
                             output=output)
        self.assertEqual(
            {'status': 'FAILED',
             'error': "Node 'NODE_ID' is attached to more than one "
                       "network. Please pick the network to use."},
            output)

    def test__run_script_no_network(self):
        addr = {}
        output = {}

        self.cmd._run_script('NODE_ID', addr, '', 'floating', 22, 'john',
                             False, 'identity_path', 'echo foo', '-f bar',
                             output=output)

        self.assertEqual(
            {'status': 'FAILED',
             'error': "Node 'NODE_ID' is not attached to any network."},
            output)

    def test__run_script_no_matching_address(self):
        addr = {
            'private': [
                {
                    'OS-EXT-IPS:type': 'fixed',
                    'version': 4,
                    'addr': '1.2.3.4',
                }
            ]
        }
        output = {}

        self.cmd._run_script('NODE_ID', addr, 'private', 'floating', 22,
                             'john', False, 'identity_path', 'echo foo',
                             '-f bar',
                             output=output)
        self.assertEqual(
            {'status': 'FAILED',
             'error': "No address that matches network 'private' and "
                      "type 'floating' of IPv4 has been found for node "
                      "'NODE_ID'."},
            output)

    def test__run_script_more_than_one_address(self):
        addr = {
            'private': [
                {
                    'OS-EXT-IPS:type': 'fixed',
                    'version': 4,
                    'addr': '1.2.3.4',
                },
                {
                    'OS-EXT-IPS:type': 'fixed',
                    'version': 4,
                    'addr': '5.6.7.8',
                },
            ]
        }

        output = {}

        self.cmd._run_script('NODE_ID', addr, 'private', 'fixed', 22, 'john',
                             False, 'identity_path', 'echo foo', '-f bar',
                             output=output)
        self.assertEqual(
            {'status': 'FAILED',
             'error': "More than one IPv4 fixed address found."},
            output)

    @mock.patch('threading.Thread')
    @mock.patch.object(osc_cluster.ClusterRun, '_run_script')
    def test_cluster_run(self, mock_script, mock_thread):
        arglist = [
            '--port', '22',
            '--address-type', 'fixed',
            '--network', 'private',
            '--user', 'root',
            '--identity-file', 'path-to-identity',
            '--ssh-options', '-f boo',
            '--script', 'script-file',
            'cluster1'
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])

        th1 = mock.Mock()
        th2 = mock.Mock()
        mock_thread.side_effect = [th1, th2]
        fake_script = 'blah blah'
        with mock.patch('senlinclient.v1.cluster.open',
                        mock.mock_open(read_data=fake_script)) as mock_open:
            self.cmd.take_action(parsed_args)

        self.mock_client.collect_cluster_attrs.assert_called_once_with(
            'cluster1', 'details')
        mock_open.assert_called_once_with('script-file', 'r')
        mock_thread.assert_has_calls([
            mock.call(target=mock_script,
                      args=('NODE_ID1', 'ADDRESS CONTENT 1', 'private',
                            'fixed', 22, 'root', False, 'path-to-identity',
                            'blah blah', '-f boo'),
                      kwargs={'output': {}}),
            mock.call(target=mock_script,
                      args=('NODE_ID2', 'ADDRESS CONTENT 2', 'private',
                            'fixed', 22, 'root', False, 'path-to-identity',
                            'blah blah', '-f boo'),
                      kwargs={'output': {}})
        ])
        th1.start.assert_called_once_with()
        th2.start.assert_called_once_with()
        th1.join.assert_called_once_with()
        th2.join.assert_called_once_with()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import time

from oslo_utils import uuidutils
from tempest.lib.cli import base
from tempest.lib.cli import output_parser
from tempest.lib import exceptions as tempest_lib_exc


class OpenStackClientTestBase(base.ClientTestBase):
    """Command line client base functions."""

    def setUp(self):
        super(OpenStackClientTestBase, self).setUp()
        self.parser = output_parser

    def _get_clients(self):
        cli_dir = os.environ.get(
            'OS_SENLINCLIENT_EXEC_DIR',
            os.path.join(os.path.abspath('.'), '.tox/functional/bin'))

        return base.CLIClient(
            username=os.environ.get('OS_USERNAME'),
            password=os.environ.get('OS_PASSWORD'),
            tenant_name=os.environ.get('OS_PROJECT_NAME',
                                       os.environ.get('OS_TENANT_NAME')),
            uri=os.environ.get('OS_AUTH_URL'),
            cli_dir=cli_dir)

    def openstack(self, *args, **kwargs):
        return self.clients.openstack(*args, **kwargs)

    def show_to_dict(self, output):
        obj = {}
        items = self.parser.listing(output)
        for item in items:
            obj[item['Field']] = str(item['Value'])
        return dict((self._key_name(k), v) for k, v in obj.items())

    def _key_name(self, key):
        return key.lower().replace(' ', '_')

    def name_generate(self):
        """Generate randomized name for some entity."""
        name = uuidutils.generate_uuid()[:8]
        return name

    def _get_profile_path(self, profile_name):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'profiles/%s' % profile_name)

    def _get_policy_path(self, policy_name):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'policies/%s' % policy_name)

    def wait_for_status(self, name, status, check_type, timeout=60,
                        poll_interval=5):
        """Wait until name reaches given status.

        :param name: node or cluster name
        :param status: expected status of node or cluster
        :param timeout: timeout in seconds
        :param poll_interval: poll interval in seconds
        """
        if check_type == 'node':
            cmd = ('cluster node show %s' % name)
        elif check_type == 'cluster':
            cmd = ('cluster show %s' % name)
        time.sleep(poll_interval)
        start_time = time.time()
        while time.time() - start_time < timeout:
            check_status = self.openstack(cmd)
            result = self.show_to_dict(check_status)
            if result['status'] == status:
                break
            time.sleep(poll_interval)
        else:
            message = ("%s %s did not reach status %s after %d s"
                       % (check_type, name, status, timeout))
            raise tempest_lib_exc.TimeoutException(message)

    def wait_for_delete(self, name, check_type, timeout=60,
                        poll_interval=5):
        """Wait until delete finish"""
        if check_type == 'node':
            cmd = ('cluster node show %s' % name)
        if check_type == 'cluster':
            cmd = ('cluster show %s' % name)
        time.sleep(poll_interval)
        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                self.openstack(cmd)
            except tempest_lib_exc.CommandFailed as ex:
                if "No Node found" or "No Cluster found" in ex.stderr:
                    break
            time.sleep(poll_interval)
        else:

            message = ("failed in deleting %s %s after %d seconds"
                       % (check_type, name, timeout))
            raise tempest_lib_exc.TimeoutException(message)

    def policy_create(self, name, policy='deletion_policy.yaml'):
        pf = self._get_policy_path(policy)
        cmd = ('cluster policy create --spec-file %s %s'
               % (pf, name))
        policy_raw = self.openstack(cmd)
        result = self.show_to_dict(policy_raw)
        return result

    def policy_delete(self, name_or_id):
        cmd = ('cluster policy delete %s --force' % name_or_id)
        self.openstack(cmd)

    def profile_create(self, name, profile='cirros_basic.yaml'):
        pf = self._get_profile_path(profile)
        cmd = ('cluster profile create --spec-file %s %s'
               % (pf, name))
        profile_raw = self.openstack(cmd)
        result = self.show_to_dict(profile_raw)
        return result

    def profile_delete(self, name_or_id):
        cmd = ('cluster profile delete %s --force' % name_or_id)
        self.openstack(cmd)

    def node_create(self, profile, name):
        cmd = ('cluster node create --profile %s %s'
               % (profile, name))
        node_raw = self.openstack(cmd)
        result = self.show_to_dict(node_raw)
        self.wait_for_status(name, 'ACTIVE', 'node', 120)
        return result

    def node_delete(self, name_or_id):
        cmd = ('cluster node delete %s --force' % name_or_id)
        self.openstack(cmd)
        self.wait_for_delete(name_or_id, 'node', 120)

    def cluster_create(self, profile, name, desired_capacity=0):
        cmd = ('cluster create --profile %s --desired-capacity %d %s'
               % (profile, desired_capacity, name))
        cluster_raw = self.openstack(cmd)
        result = self.show_to_dict(cluster_raw)
        self.wait_for_status(name, 'ACTIVE', 'cluster', 120)
        return result

    def cluster_delete(self, name_or_id):
        cmd = ('cluster delete %s --force' % name_or_id)
        self.openstack(cmd)
        self.wait_for_delete(name_or_id, 'cluster', 120)

    def receiver_create(self, name, cluster, action='CLUSTER_SCALE_OUT',
                        rt='webhook'):
        cmd = ('cluster receiver create --cluster %s --action %s --type %s '
               '%s' % (cluster, action, rt, name))
        receiver_raw = self.openstack(cmd)
        result = self.show_to_dict(receiver_raw)
        return result

    def receiver_delete(self, name_or_id):
        cmd = ('cluster receiver delete %s --force' % name_or_id)
        self.openstack(cmd)
